Class: ActiveStorage::Blob
- Includes:
- Analyzable, Identifiable, Representable, Servable
- Defined in:
- app/models/active_storage/blob.rb
Overview
Active Storage Blob
A blob is a record that contains the metadata about a file and a key for where that file resides on the service. Blobs can be created in two ways:
-
Ahead of the file being uploaded server-side to the service, via
create_and_upload!. A rewindableiowith the file contents must be available at the server for this operation. -
Ahead of the file being directly uploaded client-side to the service, via
create_before_direct_upload!.
The first option doesn’t require any client-side JavaScript integration, and can be used by any other back-end service that deals with files. The second option is faster, since you’re not using your own server as a staging point for uploads, and can work with deployments like Heroku that do not provide large amounts of disk space.
Blobs are intended to be immutable in as-so-far as their reference to a specific file goes. You’re allowed to update a blob’s metadata on a subsequent pass, but you should not update the key or change the uploaded file. If you need to create a derivative or otherwise change the blob, simply create a new blob and purge the old one.
When using a custom key, the value is treated as trusted. Using untrusted user input as the key may result in unexpected behavior.
Defined Under Namespace
Modules: Analyzable, Identifiable, Representable, Servable
Constant Summary collapse
- MINIMUM_TOKEN_LENGTH =
28
Class Method Summary collapse
-
.build_after_unfurling(key: nil, io:, filename:, content_type: nil, metadata: nil, service_name: nil, identify: true, record: nil) ⇒ Object
:nodoc:.
-
.combine_signed_id_purposes(purpose) ⇒ Object
Customize signed ID purposes for backwards compatibility.
-
.compose(blobs, key: nil, filename:, content_type: nil, metadata: nil) ⇒ Object
Concatenate multiple blobs into a single “composed” blob.
-
.create_after_unfurling!(key: nil, io:, filename:, content_type: nil, metadata: nil, service_name: nil, identify: true, record: nil) ⇒ Object
:nodoc:.
-
.create_and_upload!(key: nil, io:, filename:, content_type: nil, metadata: nil, service_name: nil, identify: true, record: nil) ⇒ Object
Creates a new blob instance and then uploads the contents of the given
ioto the service. -
.create_before_direct_upload!(key: nil, filename:, byte_size:, checksum:, content_type: nil, metadata: nil, service_name: nil, record: nil) ⇒ Object
Returns a saved blob without uploading a file to the service.
-
.find_signed(id, record: nil, purpose: :blob_id) ⇒ Object
You can use the signed ID of a blob to refer to it on the client side without fear of tampering.
-
.find_signed!(id, record: nil, purpose: :blob_id) ⇒ Object
Works like
find_signed, but will raise anActiveSupport::MessageVerifier::InvalidSignatureexception if thesigned_idhas either expired, has a purpose mismatch, or has been tampered with. -
.generate_unique_secure_token(length: MINIMUM_TOKEN_LENGTH) ⇒ Object
To prevent problems with case-insensitive filesystems, especially in combination with databases which treat indices as case-sensitive, all blob keys generated are going to only contain the base-36 character alphabet and will therefore be lowercase.
-
.scope_for_strict_loading ⇒ Object
:nodoc:.
-
.signed_id_verifier ⇒ Object
Customize the default signed ID verifier for backwards compatibility.
Instance Method Summary collapse
-
#attachments ⇒ Object
:method:.
-
#audio? ⇒ Boolean
Returns true if the content_type of this blob is in the audio range, like audio/mpeg.
-
#compose(keys) ⇒ Object
:nodoc:.
- #custom_metadata ⇒ Object
- #custom_metadata=(metadata) ⇒ Object
-
#delete ⇒ Object
Deletes the files on the service associated with the blob.
-
#download(&block) ⇒ Object
Downloads the file associated with this blob.
-
#download_chunk(range) ⇒ Object
Downloads a part of the file associated with this blob.
-
#filename ⇒ Object
Returns an ActiveStorage::Filename instance of the filename that can be queried for basename, extension, and a sanitized version of the filename that’s safe to use in URLs.
-
#image? ⇒ Boolean
Returns true if the content_type of this blob is in the image range, like image/png.
-
#key ⇒ Object
Returns the key pointing to the file on the service that’s associated with this blob.
-
#mirror_later ⇒ Object
:nodoc:.
-
#open(tmpdir: nil, &block) ⇒ Object
Downloads the blob to a tempfile on disk.
-
#purge ⇒ Object
Destroys the blob record and then deletes the file on the service.
-
#purge_later ⇒ Object
Enqueues an ActiveStorage::PurgeJob to call #purge.
-
#service ⇒ Object
Returns an instance of service, which can be configured globally or per attachment.
-
#service_headers_for_direct_upload ⇒ Object
Returns a Hash of headers for
service_url_for_direct_uploadrequests. -
#service_url_for_direct_upload(expires_in: ActiveStorage.service_urls_expire_in) ⇒ Object
Returns a URL that can be used to directly upload a file for this blob on the service.
-
#signed_id(purpose: :blob_id, expires_in: nil, expires_at: nil) ⇒ Object
Returns a signed ID for this blob that’s suitable for reference on the client-side without fear of tampering.
-
#text? ⇒ Boolean
Returns true if the content_type of this blob is in the text range, like text/plain.
-
#unattached ⇒ Object
:singleton-method:.
-
#unfurl(io, identify: true) ⇒ Object
:nodoc:.
-
#upload(io, identify: true) ⇒ Object
Uploads the
ioto the service on thekeyfor this blob. -
#upload_without_unfurling(io) ⇒ Object
:nodoc:.
-
#url(expires_in: ActiveStorage.service_urls_expire_in, disposition: :inline, filename: nil, **options) ⇒ Object
Returns the URL of the blob on the service.
-
#video? ⇒ Boolean
Returns true if the content_type of this blob is in the video range, like video/mp4.
Methods included from Servable
#content_type_for_serving, #forced_disposition_for_serving
Methods included from Representable
#create_preview_image_later, #preprocessed, #preview, #preview_image_needed_before_processing_variants?, #previewable?, #representable?, #representation, #variable?, #variant
Methods included from Identifiable
#identified?, #identify, #identify_without_saving
Methods included from Analyzable
#analyze, #analyze_later, #analyzed?
Class Method Details
.build_after_unfurling(key: nil, io:, filename:, content_type: nil, metadata: nil, service_name: nil, identify: true, record: nil) ⇒ Object
:nodoc:
88 89 90 91 92 |
# File 'app/models/active_storage/blob.rb', line 88 def build_after_unfurling(key: nil, io:, filename:, content_type: nil, metadata: nil, service_name: nil, identify: true, record: nil) # :nodoc: new(key: key, filename: filename, content_type: content_type, metadata: , service_name: service_name).tap do |blob| blob.unfurl(io, identify: identify) end end |
.combine_signed_id_purposes(purpose) ⇒ Object
Customize signed ID purposes for backwards compatibility.
132 133 134 |
# File 'app/models/active_storage/blob.rb', line 132 def combine_signed_id_purposes(purpose) # :nodoc: purpose.to_s end |
.compose(blobs, key: nil, filename:, content_type: nil, metadata: nil) ⇒ Object
Concatenate multiple blobs into a single “composed” blob.
156 157 158 159 160 161 162 163 164 165 |
# File 'app/models/active_storage/blob.rb', line 156 def compose(blobs, key: nil, filename:, content_type: nil, metadata: nil) raise ActiveRecord::RecordNotSaved, "All blobs must be persisted." if blobs.any?(&:new_record?) content_type ||= blobs.pluck(:content_type).compact.first new(key: key, filename: filename, content_type: content_type, metadata: , byte_size: blobs.sum(&:byte_size)).tap do |combined_blob| combined_blob.compose(blobs.pluck(:key)) combined_blob.save! end end |
.create_after_unfurling!(key: nil, io:, filename:, content_type: nil, metadata: nil, service_name: nil, identify: true, record: nil) ⇒ Object
:nodoc:
94 95 96 |
# File 'app/models/active_storage/blob.rb', line 94 def create_after_unfurling!(key: nil, io:, filename:, content_type: nil, metadata: nil, service_name: nil, identify: true, record: nil) # :nodoc: build_after_unfurling(key: key, io: io, filename: filename, content_type: content_type, metadata: , service_name: service_name, identify: identify).tap(&:save!) end |
.create_and_upload!(key: nil, io:, filename:, content_type: nil, metadata: nil, service_name: nil, identify: true, record: nil) ⇒ Object
Creates a new blob instance and then uploads the contents of the given io to the service. The blob instance is going to be saved before the upload begins to prevent the upload clobbering another due to key collisions. When providing a content type, pass identify: false to bypass automatic content type inference.
The optional key parameter is treated as trusted. Using untrusted user input as the key may result in unexpected behavior.
106 107 108 109 110 |
# File 'app/models/active_storage/blob.rb', line 106 def create_and_upload!(key: nil, io:, filename:, content_type: nil, metadata: nil, service_name: nil, identify: true, record: nil) create_after_unfurling!(key: key, io: io, filename: filename, content_type: content_type, metadata: , service_name: service_name, identify: identify).tap do |blob| blob.upload_without_unfurling(io) end end |
.create_before_direct_upload!(key: nil, filename:, byte_size:, checksum:, content_type: nil, metadata: nil, service_name: nil, record: nil) ⇒ Object
Returns a saved blob without uploading a file to the service. This blob will point to a key where there is no file yet. It’s intended to be used together with a client-side upload, which will first create the blob in order to produce the signed URL for uploading. This signed URL points to the key generated by the blob. Once the form using the direct upload is submitted, the blob can be associated with the right record using the signed ID.
117 118 119 120 |
# File 'app/models/active_storage/blob.rb', line 117 def create_before_direct_upload!(key: nil, filename:, byte_size:, checksum:, content_type: nil, metadata: nil, service_name: nil, record: nil) = () create! key: key, filename: filename, byte_size: byte_size, checksum: checksum, content_type: content_type, metadata: , service_name: service_name end |
.find_signed(id, record: nil, purpose: :blob_id) ⇒ Object
You can use the signed ID of a blob to refer to it on the client side without fear of tampering. This is particularly helpful for direct uploads where the client-side needs to refer to the blob that was created ahead of the upload itself on form submission.
The signed ID is also used to create stable URLs for the blob through the BlobsController.
77 78 79 |
# File 'app/models/active_storage/blob.rb', line 77 def find_signed(id, record: nil, purpose: :blob_id) super(id, purpose: purpose) end |
.find_signed!(id, record: nil, purpose: :blob_id) ⇒ Object
Works like find_signed, but will raise an ActiveSupport::MessageVerifier::InvalidSignature exception if the signed_id has either expired, has a purpose mismatch, or has been tampered with. It will also raise an ActiveRecord::RecordNotFound exception if the valid signed id can’t find a record.
84 85 86 |
# File 'app/models/active_storage/blob.rb', line 84 def find_signed!(id, record: nil, purpose: :blob_id) super(id, purpose: purpose) end |
.generate_unique_secure_token(length: MINIMUM_TOKEN_LENGTH) ⇒ Object
To prevent problems with case-insensitive filesystems, especially in combination with databases which treat indices as case-sensitive, all blob keys generated are going to only contain the base-36 character alphabet and will therefore be lowercase. To maintain the same or higher amount of entropy as in the base-58 encoding used by has_secure_token the number of bytes used is increased to 28 from the standard 24
127 128 129 |
# File 'app/models/active_storage/blob.rb', line 127 def generate_unique_secure_token(length: MINIMUM_TOKEN_LENGTH) SecureRandom.base36(length) end |
.scope_for_strict_loading ⇒ Object
:nodoc:
144 145 146 147 148 149 150 151 152 153 |
# File 'app/models/active_storage/blob.rb', line 144 def scope_for_strict_loading # :nodoc: if strict_loading_by_default? && ActiveStorage.track_variants includes( variant_records: { image_attachment: :blob }, preview_image_attachment: { blob: { variant_records: { image_attachment: :blob } } } ) else all end end |
.signed_id_verifier ⇒ Object
Customize the default signed ID verifier for backwards compatibility.
We override the reader (.signed_id_verifier) instead of just calling the writer (.signed_id_verifier=) to guard against the case where ActiveStorage.verifier isn’t yet initialized at load time.
140 141 142 |
# File 'app/models/active_storage/blob.rb', line 140 def signed_id_verifier # :nodoc: @signed_id_verifier ||= ActiveStorage.verifier end |
Instance Method Details
#attachments ⇒ Object
:method:
Returns the associated ActiveStorage::Attachment instances.
40 |
# File 'app/models/active_storage/blob.rb', line 40 has_many :attachments, autosave: false |
#audio? ⇒ Boolean
Returns true if the content_type of this blob is in the audio range, like audio/mpeg.
217 218 219 |
# File 'app/models/active_storage/blob.rb', line 217 def audio? content_type&.start_with?("audio") end |
#compose(keys) ⇒ Object
:nodoc:
280 281 282 283 |
# File 'app/models/active_storage/blob.rb', line 280 def compose(keys) # :nodoc: self.composed = true service.compose(keys, key, **) end |
#custom_metadata ⇒ Object
203 204 205 |
# File 'app/models/active_storage/blob.rb', line 203 def self[:metadata][:custom] || {} end |
#custom_metadata=(metadata) ⇒ Object
207 208 209 |
# File 'app/models/active_storage/blob.rb', line 207 def () self[:metadata] = self[:metadata].merge(custom: ) end |
#delete ⇒ Object
Deletes the files on the service associated with the blob. This should only be done if the blob is going to be deleted as well or you will essentially have a dead reference. It’s recommended to use #purge and #purge_later methods in most circumstances.
327 328 329 330 |
# File 'app/models/active_storage/blob.rb', line 327 def delete service.delete(key) service.delete_prefixed("variants/#{key}/") if image? end |
#download(&block) ⇒ Object
Downloads the file associated with this blob. If no block is given, the entire file is read into memory and returned. That’ll use a lot of RAM for very large files. If a block is given, then the download is streamed and yielded in chunks.
287 288 289 |
# File 'app/models/active_storage/blob.rb', line 287 def download(&block) service.download key, &block end |
#download_chunk(range) ⇒ Object
Downloads a part of the file associated with this blob.
292 293 294 |
# File 'app/models/active_storage/blob.rb', line 292 def download_chunk(range) service.download_chunk key, range end |
#filename ⇒ Object
Returns an ActiveStorage::Filename instance of the filename that can be queried for basename, extension, and a sanitized version of the filename that’s safe to use in URLs.
199 200 201 |
# File 'app/models/active_storage/blob.rb', line 199 def filename ActiveStorage::Filename.new(self[:filename]) end |
#image? ⇒ Boolean
Returns true if the content_type of this blob is in the image range, like image/png.
212 213 214 |
# File 'app/models/active_storage/blob.rb', line 212 def image? content_type&.start_with?("image") end |
#key ⇒ Object
Returns the key pointing to the file on the service that’s associated with this blob. The key is the secure-token format from Rails in lower case. So it’ll look like: xtapjjcjiudrlk3tmwyjgpuobabd. This key is not intended to be revealed directly to the user. Always refer to blobs using the signed_id or a verified form of the key.
191 192 193 194 |
# File 'app/models/active_storage/blob.rb', line 191 def key # We can't wait until the record is first saved to have a key for it self[:key] ||= self.class.generate_unique_secure_token(length: MINIMUM_TOKEN_LENGTH) end |
#mirror_later ⇒ Object
:nodoc:
320 321 322 |
# File 'app/models/active_storage/blob.rb', line 320 def mirror_later # :nodoc: service.mirror_later key, checksum: checksum if service.respond_to?(:mirror_later) end |
#open(tmpdir: nil, &block) ⇒ Object
Downloads the blob to a tempfile on disk. Yields the tempfile.
The tempfile’s name is prefixed with ActiveStorage- and the blob’s ID. Its extension matches that of the blob.
By default, the tempfile is created in Dir.tmpdir. Pass tmpdir: to create it in a different directory:
blob.open(tmpdir: "/path/to/tmp") do |file|
# ...
end
The tempfile is automatically closed and unlinked after the given block is executed.
Raises ActiveStorage::IntegrityError if the downloaded data does not match the blob’s checksum.
309 310 311 312 313 314 315 316 317 318 |
# File 'app/models/active_storage/blob.rb', line 309 def open(tmpdir: nil, &block) service.open( key, checksum: checksum, verify: !composed, name: [ "ActiveStorage-#{id}-", filename.extension_with_delimiter ], tmpdir: tmpdir, &block ) end |
#purge ⇒ Object
Destroys the blob record and then deletes the file on the service. This is the recommended way to dispose of unwanted blobs. Note, though, that deleting the file off the service will initiate an HTTP connection to the service, which may be slow or prevented, so you should not use this method inside a transaction or in callbacks. Use #purge_later instead.
335 336 337 338 339 |
# File 'app/models/active_storage/blob.rb', line 335 def purge destroy delete if previously_persisted? rescue ActiveRecord::InvalidForeignKey end |
#purge_later ⇒ Object
Enqueues an ActiveStorage::PurgeJob to call #purge. This is the recommended way to purge blobs from a transaction, an Active Record callback, or in any other real-time scenario.
343 344 345 |
# File 'app/models/active_storage/blob.rb', line 343 def purge_later ActiveStorage::PurgeJob.perform_later(self) end |
#service ⇒ Object
Returns an instance of service, which can be configured globally or per attachment
348 349 350 |
# File 'app/models/active_storage/blob.rb', line 348 def service services.fetch(service_name) end |
#service_headers_for_direct_upload ⇒ Object
Returns a Hash of headers for service_url_for_direct_upload requests.
247 248 249 |
# File 'app/models/active_storage/blob.rb', line 247 def service_headers_for_direct_upload service.headers_for_direct_upload key, filename: filename, content_type: content_type, content_length: byte_size, checksum: checksum, custom_metadata: end |
#service_url_for_direct_upload(expires_in: ActiveStorage.service_urls_expire_in) ⇒ Object
Returns a URL that can be used to directly upload a file for this blob on the service. This URL is intended to be short-lived for security and only generated on-demand by the client-side JavaScript responsible for doing the uploading.
242 243 244 |
# File 'app/models/active_storage/blob.rb', line 242 def service_url_for_direct_upload(expires_in: ActiveStorage.service_urls_expire_in) service.url_for_direct_upload key, expires_in: expires_in, content_type: content_type, content_length: byte_size, checksum: checksum, custom_metadata: end |
#signed_id(purpose: :blob_id, expires_in: nil, expires_at: nil) ⇒ Object
Returns a signed ID for this blob that’s suitable for reference on the client-side without fear of tampering.
183 184 185 |
# File 'app/models/active_storage/blob.rb', line 183 def signed_id(purpose: :blob_id, expires_in: nil, expires_at: nil) super end |
#text? ⇒ Boolean
Returns true if the content_type of this blob is in the text range, like text/plain.
227 228 229 |
# File 'app/models/active_storage/blob.rb', line 227 def text? content_type&.start_with?("text") end |
#unattached ⇒ Object
:singleton-method:
Returns the blobs that aren’t attached to any record.
46 |
# File 'app/models/active_storage/blob.rb', line 46 scope :unattached, -> { where.missing(:attachments) } |
#unfurl(io, identify: true) ⇒ Object
:nodoc:
269 270 271 272 273 274 |
# File 'app/models/active_storage/blob.rb', line 269 def unfurl(io, identify: true) # :nodoc: self.checksum = compute_checksum_in_chunks(io) self.content_type = extract_content_type(io) if content_type.nil? || identify self.byte_size = io.size self.identified = true end |
#upload(io, identify: true) ⇒ Object
Uploads the io to the service on the key for this blob. Blobs are intended to be immutable, so you shouldn’t be using this method after a file has already been uploaded to fit with a blob. If you want to create a derivative blob, you should instead simply create a new blob based on the old one.
Prior to uploading, we compute the checksum, which is sent to the service for transit integrity validation. If the checksum does not match what the service receives, an exception will be raised. We also measure the size of the io and store that in byte_size on the blob record. The content type is automatically extracted from the io unless you specify a content_type and pass identify as false.
Normally, you do not have to call this method directly at all. Use the create_and_upload! class method instead. If you do use this method directly, make sure you are using it on a persisted Blob as otherwise another blob’s data might get overwritten on the service.
264 265 266 267 |
# File 'app/models/active_storage/blob.rb', line 264 def upload(io, identify: true) unfurl io, identify: identify upload_without_unfurling io end |
#upload_without_unfurling(io) ⇒ Object
:nodoc:
276 277 278 |
# File 'app/models/active_storage/blob.rb', line 276 def upload_without_unfurling(io) # :nodoc: service.upload key, io, checksum: checksum, ** end |
#url(expires_in: ActiveStorage.service_urls_expire_in, disposition: :inline, filename: nil, **options) ⇒ Object
Returns the URL of the blob on the service. This returns a permanent URL for public files, and returns a short-lived URL for private files. Private files are signed, and not for public use. Instead, the URL should only be exposed as a redirect from a stable, possibly authenticated URL. Hiding the URL behind a redirect also allows you to change services without updating all URLs.
235 236 237 238 |
# File 'app/models/active_storage/blob.rb', line 235 def url(expires_in: ActiveStorage.service_urls_expire_in, disposition: :inline, filename: nil, **) service.url key, expires_in: expires_in, filename: ActiveStorage::Filename.wrap(filename || self.filename), content_type: content_type_for_serving, disposition: forced_disposition_for_serving || disposition, ** end |
#video? ⇒ Boolean
Returns true if the content_type of this blob is in the video range, like video/mp4.
222 223 224 |
# File 'app/models/active_storage/blob.rb', line 222 def video? content_type&.start_with?("video") end |