Class: ActiveStorage::DiskController

Inherits:
BaseController
  • Object
show all
Includes:
FileServer
Defined in:
app/controllers/active_storage/disk_controller.rb

Overview

Serves files stored with the disk service in the same way that the cloud services do. This means using expiring, signed URLs that are meant for immediate access, not permanent linking. Always go through the BlobsController, or your own authenticated controller, rather than directly to the service URL.

Instance Method Summary collapse

Instance Method Details

#showObject



12
13
14
15
16
17
18
19
20
21
22
# File 'app/controllers/active_storage/disk_controller.rb', line 12

def show
  if key = decode_verified_key
    serve_file named_disk_service(key[:service_name]).path_for(key[:key]), content_type: key[:content_type], disposition: key[:disposition]
  else
    head :not_found
  end
rescue Errno::ENOENT
  head :not_found
rescue ActiveStorage::InvalidKeyError
  head :not_found
end

#updateObject



24
25
26
27
28
29
30
31
32
33
34
35
36
37
38
39
# File 'app/controllers/active_storage/disk_controller.rb', line 24

def update
  if token = decode_verified_token
    if acceptable_content?(token)
      named_disk_service(token[:service_name]).upload token[:key], request.body, checksum: token[:checksum]
      head :no_content
    else
      head ActionDispatch::Constants::UNPROCESSABLE_CONTENT
    end
  else
    head :not_found
  end
rescue ActiveStorage::IntegrityError
  head ActionDispatch::Constants::UNPROCESSABLE_CONTENT
rescue ActiveStorage::InvalidKeyError
  head ActionDispatch::Constants::UNPROCESSABLE_CONTENT
end