Exception: ActiveStorage::InvalidKeyError
- Defined in:
- lib/active_storage/errors.rb
Overview
Raised when a storage key resolves to a path outside the service’s root directory, indicating a potential path traversal attack.
Raised when a storage key resolves to a path outside the service’s root directory, indicating a potential path traversal attack.